/*
 * This is an example of the Tcl/C++ interface.
 *
 * It's a very tiny demonstration of abstract base classes and virtual
 * functions.
 *
 * We define some trivial "storage" class. We have a couple of classes
 * that can store a value. The only thing they have in common is that
 * the value can be queried as a string.
 *
 * The base class therefore defines the pure virtual function query,
 * which is then overloaded by the storage classes for an integer,
 * a double and a string. Each derived class also defines a "set"
 * function.
 *
 * On the Tcl level, however, we provide a query function for the
 * base class, which on the C++ level automatically calls the
 * correct derived function.
 * The Tcl level instances of derived functions inherit the base
 * class's query function.
 */

#ifndef __ABSTRACT_H__
#define __ABSTRACT_H__

#ifdef HAVE_TCLOBJ
#include <tcl.h>
#include <tclobj.h>
#endif

/*
 * base class for general storage
 */

#ifdef HAVE_TCLOBJ
class storage : public Tcl_Object
#else
class storage
#endif
{
public:
  virtual char * query (void) = 0;

#ifdef HAVE_TCLOBJ
  int Tcl_Proc (Tcl_Interp *, int, char *[], int);
  TCL_OBJECT_DECL_CLASS(storage);
#endif
};

/*
 * store an integer
 */

class int_storage : public storage
{
public:
  void   set   (int);
  char * query (void);

#ifdef HAVE_TCLOBJ
  int_storage  (Tcl_Interp *, int, char *[], int);
  int Tcl_Proc (Tcl_Interp *, int, char *[], int);
  TCL_OBJECT_DECL_CLASS(int_storage);
#endif

private:
  int value;
};

/*
 * store a double value
 */

class double_storage : public storage
{
public:
  void   set   (double);
  char * query (void);

#ifdef HAVE_TCLOBJ
  double_storage (Tcl_Interp *, int, char *[], int);
  int Tcl_Proc   (Tcl_Interp *, int, char *[], int);
  TCL_OBJECT_DECL_CLASS(double_storage);
#endif

private:
  double value;
};

/*
 * store a string
 */

class string_storage : public storage
{
public:
  string_storage (void);
  void   set   (const char *);
  char * query (void);

#ifdef HAVE_TCLOBJ
  string_storage (Tcl_Interp *, int, char *[], int);
  int Tcl_Proc   (Tcl_Interp *, int, char *[], int);
  TCL_OBJECT_DECL_CLASS(string_storage);
#endif

private:
  char string[1024];
};

#endif
