/*
 * These functions interface the C++ objects with the Tcl level.
 *
 * We add some TCL commands of our own:
 *
 *  init class ?filename?
 *      Make "class" available from the Tcl level. First, look for its
 *      signature in the main executable and the already loaded shlibs,
 *      otherwise try to resolve it from filename.
 *
 *  new class ?args ...?
 *      Create a new object of the given type. Pass the constructor some
 *      arguments.
 *
 *  delete object ?object ...?
 *      Destroy the given objects, calling its destructor.
 *
 *
 * We export some C++ helper functions as well.
 *
 *  Tcl_Object * TclObj_Cast (Tcl_Interp * interp,
 *                            const char * objname,
 *                            const char * thetype)
 *
 *      Check if the argument by name objname is compatible with the
 *      type thetype. Return a pointer to the object if true, or null
 *      if they are incompatible, leaving an error message in interp.
 *
 *  Tcl_Object * TclObj_Cast (Tcl_Interp * interp,
 *                            const Tcl_Object * me,
 *                            unsigned long oident)
 *
 *      Same as above, but with different argument types.
 *
 *  TclObj_Info * TclObj_Require (Tcl_Interp * interp,
 *                                const char * classname)
 *
 *      Require that the class classname is available.
 *
 *  int TclObj_Register (Tcl_Interp * interp,
 *                       TclObj_Info * info)
 *
 *      Register a class from it's signature.
 *
 *  int TclObj_GetLong         (Tcl_Interp *, const char *, long *)
 *  int TclObj_GetUnsignedLong (Tcl_Interp *, const char *, unsigned long *)
 *  int TclObj_GetUnsignedInt  (Tcl_Interp *, const char *, unsigned int *)
 *
 *      Similar to the Tcl_Get* functions, but with different argument
 *      types. Tcl_GetInt does *only* work for integers.
 *
 *  int TclObj_CastArgs (Tcl_Interp *, int, char *[], const char *, ...)
 *
 *      Cast a lot of arguments, similar to a sscanf() statement. The
 *      parameter list is given in the same way as in the command tables.
 *
 *  int TclObj_MkResult (Tcl_Interp *, const char *, ...)
 *
 *      The reverse. Produce a list of return values, similar to a printf()
 *      statement, but again the parameter list is like in a command table.
 *
 * The functions that interface with Tcl must be of C linkage.
 * The helper functions should be C++ linkage.
 */

#include <tcl.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include "tclobj.h"
#include "register.h"
#include "compat.h"
#include "signature.h"
#include "fptools.h"

/*
 * cvs version
 */
char * tclobj_interface_id = "$Id: interface.cc,v 1.3 1996/12/04 08:54:32 fp Exp $";


/*
 * forward declaration of Tcl-level functions. The helper functions
 * should be declared in tclobj.h
 */

extern "C" {
static int	TclObj_class	(ClientData, Tcl_Interp *,
				 int, char *[]);
static int	TclObj_new	(ClientData, Tcl_Interp *,
				 int, char *[]);
static char *	TclObj_scope	(ClientData, Tcl_Interp *,
				 char *, char *, int);
static int	TclObj_delete	(ClientData, Tcl_Interp *,
				 int, char *[]);
}

/*
 * ----------------------------------------------------------------------
 * C++-Level part. These functions should be of C++ linkage
 * ----------------------------------------------------------------------
 */

/*
 * Type Checker
 *
 * Resolve command name and return it as an object if it matches the
 * given type, or NULL otherwise
 */

Tcl_Object *
TclObj_Cast (Tcl_Interp * interp,
	     Tcl_Object * me,
	     unsigned long oident)
{
  if (TclObj_IsCompatible (interp, me->classinfo(), oident))
    return me;
  else
    return NULL;
}

Tcl_Object *
TclObj_Cast (Tcl_Interp *interp,
	     const char * objname,
	     const char * thetype)
{
  Tcl_CmdInfo cmdinfo;
  Tcl_Object *theobj;
  TclObj_Info *theinfo;

  if (Tcl_GetCommandInfo (interp, (char *) objname, &cmdinfo) == 0) {
    Tcl_AppendResult (interp, "unknown object: \"", objname, "\"", NULL);
    return NULL;
  }

  if ((theobj = (Tcl_Object *) cmdinfo.clientData) == NULL) {
    Tcl_AppendResult (interp, "Expected object of type ",
		      (thetype)?thetype:"any",
		      " but got unknown object ",
		      objname, NULL);
    return NULL;
  }

  if (thetype == NULL)
    return theobj;

  if ((theinfo = TclObj_GetClassInfo (interp, thetype)) == NULL) {
    Tcl_AppendResult (interp, objname, ": cast from type \"",
		      theobj->classinfo()->name,
		      "\" to unknown type \"", thetype,
		      "\" failed", NULL);
    return NULL;
  }

  if (TclObj_Cast (interp, theobj, theinfo->ident) == NULL) {
    Tcl_AppendResult (interp, objname, ": expected type ", thetype,
		      " but got ", theobj->classinfo()->name, NULL);
    return NULL;
  }

  return theobj;
}

/*
 * Require that a class is loaded. Check if the class is available in
 * one of the already open modules. If not, try to load from the given
 * shared library.
 */

TclObj_Info *
TclObj_Require (Tcl_Interp *interp,
		const char *classname)
{
  TclObj_Info * ourinfo;

  /*
   * Check if already available
   */

  if ((ourinfo = TclObj_GetClassInfo (interp, classname)) != NULL) {
    return (TclObj_Info *) ourinfo;
  }

  /*
   * Before, here was code trying to perform some dynamic loading
   * from a guessed filename by itself. This was removed as not
   * necessary, with Tcl's package management being superior.
   */

  Tcl_AppendResult (interp, "class \"", (char*) classname,
		    "\" is not available.", NULL);

  return NULL;
}

/*
 * Register class with the information we have
 */

int
TclObj_Register (Tcl_Interp *interp,
		 TclObj_Info *info)
{
  int res;

  if (info == NULL || info->name == NULL) {
    Tcl_SetResult (interp, "TclObj_register: oops, called with NULL info",
		   TCL_STATIC);
    return TCL_ERROR;
  }

  /*
   * Check if already registered
   */
  if (TclObj_GetClassInfo (interp, info->name) != NULL) {
    return TCL_OK;
  }

  /*
   * Initialize all classes that this one depends on
   */

  if (info->parent) {
    if ((TclObj_Require (interp, info->parent)) == NULL)
      return TCL_ERROR;
  }

  /*
   * Register Class (generating an identifier)
   */

  TclObj_RegisterClass (interp, info);

  /*
   * If there is an initialization procedure, call it
   */

  if (info->c_init_proc) {
    if ((res = (*(info->c_init_proc)) (interp)) != TCL_OK) {
      /*
       * oops, it failed!
       */
      /*
       * we should unregister here!
       */
      return res;
    }
  }

  /*
   * install this class's name as a command to get info about the
   * class or to create objects of that type (more or less an alias
   * for new). Give info structure as clientData.
   */

  Tcl_CreateCommand (interp, (char *) info->name,
		     TclObj_class, (ClientData) info,
		     NULL);

  return TCL_OK;
}

/*
 * ----------------------------------------------------------------------
 * Get functions for string conversions to various types. Tcl only
 * provides GetInt and GetDouble. But GetInt has its problems on
 * 64bit machines, so we overload that one, too.
 * ----------------------------------------------------------------------
 */

int
TclObj_GetInt (Tcl_Interp *interp, const char *arg, int *result)
{
  long value;
  char *endp;

  errno = 0;
  value = strtol (arg, &endp, 0);

  if (endp == arg || (endp && *endp!='\0' && !isspace(*endp))) {
    if (interp) {
      Tcl_AppendResult (interp, "expected integer but got \"",
			arg, "\"", NULL);
    }
    return TCL_ERROR;
  }

  *result = (int) value;

  if (errno == ERANGE || value != (long) *result) {
    if (interp) {
      Tcl_AppendResult (interp, "value out of range of integer: \"",
			arg, "\"", NULL);
    }
    return TCL_ERROR;
  }
  return TCL_OK;
}

int
TclObj_GetLong (Tcl_Interp *interp, const char *arg, long *result)
{
  char *endp;

  errno   = 0;
  *result = strtol (arg, &endp, 0);

  if (endp == arg || (endp && *endp!='\0' && !isspace(*endp))) {
    if (interp) {
      Tcl_AppendResult (interp, "expected long but got \"", arg,
			"\"", NULL);
    }
    return TCL_ERROR;
  }
  if (errno == ERANGE) {
    if (interp) {
      Tcl_AppendResult (interp, "value out of range of long int: \"",
			arg, "\"", NULL);
    }
    return TCL_ERROR;
  }
  return TCL_OK;
}

int
TclObj_GetUnsignedLong (Tcl_Interp *interp, const char *arg,
			unsigned long *result)
{
  char *endp;

  errno   = 0;
  *result = strtoul (arg, &endp, 0);

  if (endp == arg || (endp && *endp!='\0' && !isspace(*endp))) {
    if (interp) {
      Tcl_AppendResult (interp, "expected unsigned long but got \"", 
			arg, "\"", NULL);
    }
    return TCL_ERROR;
  }
  if (errno == ERANGE) {
    if (interp) {
      Tcl_AppendResult (interp, "value out of range of unsigned long: \"",
			arg, "\"", NULL);
    }
    return TCL_ERROR;
  }
  return TCL_OK;
}

int
TclObj_GetUnsignedInt (Tcl_Interp *interp, const char *arg,
		       unsigned int *result)
{
  unsigned long value;
  char *endp;

  errno = 0;
  value = strtoul (arg, &endp, 0);

  if (endp == arg || (endp && *endp!='\0' && !isspace(*endp))) {
    if (interp) {
      Tcl_AppendResult (interp, "expected unsigned int but got \"",
			arg, "\"", NULL);
    }
    return TCL_ERROR;
  }

  *result = (unsigned int) value;

  if (errno == ERANGE || value != (unsigned long) *result) {
    if (interp) {
      Tcl_AppendResult (interp, "value out of range of unsigned int: \"",
			arg, "\"", NULL);
    }
    return TCL_ERROR;
  }
  return TCL_OK;
}

int
TclObj_GetPointer (Tcl_Interp *interp, const char *arg,
		   void **result)
{
  if (arg[0] != 'p' || arg[1] != 'v' || arg[2] != '!') {
    if (interp) {
      Tcl_AppendResult (interp, "expected pointer but got \"",
			arg, "\"", NULL);
    }
    return TCL_ERROR;
  }

  *result = (void *) strtoul (arg+3, NULL, 0);

  return TCL_OK;
}


/*
 * Cast an argument list to the desired types.
 */

int
TclObj_CastArgs (Tcl_Interp *interp,
		 int argc, char *argv[],
		 const char *arglist, ...)
{
  const char *token=arglist, **sptr;
  unsigned long ident, *ulptr;
  char tmpstr[64], *tptr;
  unsigned int *uiptr;
  int index=0, *iptr;
  double *dptr;
  va_list ap;
  void **vptr;
  long *lptr;

  va_start (ap, arglist);

  while (index<argc && *token) {
    tptr = tmpstr;
    while (isspace (*token))                 token++;
    while (*token == '.' || *token == '*')   *tptr++ = *token++;
    while (*token == '_' || isalpha(*token)) *tptr++ = *token++;
    while (*token == '_' || isalnum(*token)) *tptr++ = *token++;
    *tptr = '\0';
    while (isspace (*token)) token++;
    if (*token==',') token++;
    while (isspace (*token)) token++;

    ident = TclObj_GetClassIdent (interp, tmpstr);

    switch (ident) {
    case TCL_OBJECT_ARGTYPE_STRING:
      sptr  = va_arg (ap, const char **);
      *sptr = argv[index];
      break;

    case TCL_OBJECT_ARGTYPE_BOOL:
      iptr = va_arg (ap, int *);
      Tcl_GetBoolean (interp, (char *) argv[index], iptr);
      break;

    case TCL_OBJECT_ARGTYPE_DOUBLE:
      dptr = va_arg (ap, double *);
      Tcl_GetDouble (interp, (char *) argv[index], dptr);
      break;

    case TCL_OBJECT_ARGTYPE_INT:
      iptr = va_arg (ap, int *);
      TclObj_GetInt (interp, argv[index], iptr);
      break;

    case TCL_OBJECT_ARGTYPE_UINT:
      uiptr = va_arg (ap, unsigned int *);
      TclObj_GetUnsignedInt (interp, argv[index], uiptr);
      break;

    case TCL_OBJECT_ARGTYPE_LONG:
      lptr = va_arg (ap, long *);
      TclObj_GetLong (interp, argv[index], lptr);
      break;

    case TCL_OBJECT_ARGTYPE_ULONG:
      ulptr = va_arg (ap, unsigned long *);
      TclObj_GetUnsignedLong (interp, argv[index], ulptr);
      break;

    case TCL_OBJECT_ARGTYPE_PTR:
      vptr  = va_arg (ap, void **);
      TclObj_GetPointer (interp, argv[index], vptr);
      break;

    default: /* some object */
      vptr  = va_arg (ap, void **);
      *vptr = (void*) TclObj_Cast (interp, argv[index], tmpstr);
      break;
    }
    index++;
  }
  va_end (ap);
  return TCL_OK;
}

/*
 * Return an argument list
 */

int
TclObj_MkResult (Tcl_Interp *interp, const char *arglist, ...)
{
  char toktmp[64], tmpstr[32], *tptr;
  const char *token=arglist, *sptr;
  unsigned long ulvalue, ident;
  unsigned int uivalue;
  int ivalue; double dvalue;
  const char *thename;
  Tcl_Object *theob;
  void *theptr;
  long lvalue;
  va_list ap;

  va_start (ap, arglist);

  while (*token) {
    tptr = toktmp;
    while (isspace (*token))                 token++;
    while (*token == '.' || *token == '*')   *tptr++ = *token++;
    while (*token == '_' || isalpha(*token)) *tptr++ = *token++;
    while (*token == '_' || isalnum(*token)) *tptr++ = *token++;
    *tptr = '\0';
    while (isspace (*token)) token++;
    if (*token==',') token++;
    while (isspace (*token)) token++;

    ident = TclObj_GetClassIdent (interp, toktmp);

    switch (ident) {
    case TCL_OBJECT_ARGTYPE_STRING:
      sptr = va_arg (ap, const char *);
      Tcl_AppendElement (interp, (char *) sptr);
      break;

    case TCL_OBJECT_ARGTYPE_BOOL:
      ivalue = va_arg (ap, int);
      sprintf (tmpstr, "%d", ivalue);
      Tcl_AppendElement (interp, tmpstr);
      break;

    case TCL_OBJECT_ARGTYPE_DOUBLE:
      dvalue = va_arg (ap, double);
      sprintf (tmpstr, "%f", dvalue);
      Tcl_AppendElement (interp, tmpstr);
      break;

    case TCL_OBJECT_ARGTYPE_INT:
      ivalue = va_arg (ap, int);
      sprintf (tmpstr, "%d", ivalue);
      Tcl_AppendElement (interp, tmpstr);
      break;

    case TCL_OBJECT_ARGTYPE_UINT:
      uivalue = va_arg (ap, unsigned int);
      sprintf (tmpstr, "%u", uivalue);
      Tcl_AppendElement (interp, tmpstr);
      break;

    case TCL_OBJECT_ARGTYPE_LONG:
      lvalue = va_arg (ap, long);
      sprintf (tmpstr, "%ld", lvalue);
      Tcl_AppendElement (interp, tmpstr);
      break;

    case TCL_OBJECT_ARGTYPE_ULONG:
      ulvalue = va_arg (ap, unsigned long);
      sprintf (tmpstr, "%lu", ulvalue);
      Tcl_AppendElement (interp, tmpstr);
      break;

    case TCL_OBJECT_ARGTYPE_PTR:
      theptr = va_arg (ap, void *);
      sprintf (tmpstr, "pv!%lu", (unsigned long) theptr);
      Tcl_AppendElement (interp, tmpstr);
      break;

    default: /* some object */
      theob   = va_arg (ap, Tcl_Object *);
      thename = theob->Tcl_Name (interp);
      Tcl_AppendElement (interp, (char *) thename);
      break;
    }
  }

  va_end (ap);
  return TCL_OK;
}

/*
 * ----------------------------------------------------------------------
 * Tcl-Level part. These functions must be of C linkage
 * ----------------------------------------------------------------------
 */

extern "C" {

/*
 * Each class's name points as function to this procedure. We can either
 * provide information about the class (with the "info" option), but for
 * non-abstract classes, we can also use this C++-like syntax to create
 * objects of this type, in which case we simply refer to TclObj_new
 * We have the info structure as clientData.
 */

static int
TclObj_class (ClientData clientData, Tcl_Interp *interp,
	      int argc, char *argv[])
{
  TclObj_Info *theinfo = (TclObj_Info *) clientData;
  int index;

  if (argc<2) {
    /*
     * no parameters, print generic info
     */
    if (theinfo->parent) {
      Tcl_AppendResult  (interp, "\"", argv[0], "\": ",
			 (theinfo->c_new_proc)?"class":"abstract class",
			 " derived from ", theinfo->parent, NULL);
    }
    else {
      /*
       * no parents? this must be Tcl_Object
       */
      Tcl_AppendResult (interp, "I think you know the answer to this.",
			NULL);
    }

    return TCL_OK;
  }
  else if (_FP_stricmp (argv[1], "info") == 0) {
    if (argc == 2) {
      Tcl_AppendResult (interp, "wrong # args: should be \"",
			argv[0], " info {dep|op|sig|<func>}\"",
			NULL);
      return TCL_OK;
    }
    if ((TclObj_DumpMemberNames (interp, theinfo, argv[2], &index))==TCL_ERROR)
      return TCL_ERROR;

    if (index) {
      return TclObj_DumpSignatures (interp, theinfo, argv[2], NULL);
    }
    else if (_FP_strnicmp (argv[2], "dep", 3) == 0) {
      return TclObj_DumpComptree (interp, theinfo);
    }
    else if (_FP_strnicmp (argv[2], "op", 2) == 0) {
      return TclObj_DumpMemberNames (interp, theinfo, NULL, NULL);
    }
    else if (_FP_strnicmp (argv[2], "sig", 2) == 0) {
      return TclObj_DumpSignatures (interp, theinfo, NULL, NULL);
    }
    else {
      Tcl_AppendResult (interp, argv[0],
			" does not support ",
			argv[2], NULL);
      return TCL_ERROR;
    }
    return TCL_OK;
  }
  /*
   * any other name than info. so act as a new operator.
   * but first make sure that we have a constructor
   */
  if (theinfo->c_new_proc == NULL) {
    Tcl_AppendResult (interp, "\"", argv[0], 
		      "\" is an abstract class",
		      NULL);
    return TCL_ERROR;
  }

  /*
   * hand over to new handler
   */

  return TclObj_new (clientData, interp, argc, argv);
}

/*
 * Create new object. This procedure handles the "new" command and
 * the direct creation of objects using the name of the class (similar
 * to C++). The new command is installed with ClientData of NULL, the
 * object names with a ClientData of the class's info structure (so
 * we optimize by not looking up TclObj_Info(classname)
 */

static int
TclObj_new (ClientData clientData, Tcl_Interp *interp,
	    int argc, char *argv[])
{
  Tcl_CmdInfo cmdinfo, *oldinfo=NULL;
  const TclObj_Argtree *thetree;
  char *ptrname, *theclass;
  TclObj_Info *theinfo;
  Tcl_Object *thenew;
  int index, hadold;
  char tmpstr[256];

  if (clientData == NULL) {
    /*
     * "new" command
     */
    if (argc < 2) {
      Tcl_AppendResult (interp, "wrong # args: should be \"",
			argv[0], " class ?args ...?\"", NULL);
      return TCL_ERROR;
    }
    
    theclass = argv[1];

    /*
     * look up the info structure
     */

    if ((theinfo = TclObj_GetClassInfo (interp, theclass)) == NULL) {
      /*
       * maybe it does not exist yet? See if we can resolve it.
       */

      if ((theinfo = TclObj_Require (interp, theclass)) == NULL)
	return TCL_ERROR;
    }

    if (theinfo->c_new_proc == NULL) {
      Tcl_AppendResult (interp, "\"", theclass, 
			"\" is an abstract class",
			NULL);
      return TCL_ERROR;
    }
  }
  else {
    /*
     * "classname" desired-name
     *
     * we know that this object type is available, or we wouldn't exist.
     */
    if (argc < 2) {
      Tcl_AppendResult (interp, "wrong # args: should be \"",
			argv[0], " name ?args ...?\"", NULL);
      return TCL_ERROR;
    }

    theclass = argv[0];
    theinfo  = (TclObj_Info *) clientData;

    /*
     * Check if an object of this name does already exist. If so, we
     * "overload" the object and store the old information for later
     * perusal.
     */

    if ((hadold = Tcl_GetCommandInfo (interp, argv[1], &cmdinfo)) == 1) {
      if (cmdinfo.clientData == NULL) {
	Tcl_AppendResult (interp, "non-object of name \"", argv[1],
			  "\" does already exist", NULL);
	return TCL_ERROR;
      }
    }
  }

  index   = TCL_OK;
  thetree = TclObj_SearchSignature (interp, theinfo, "new",
				    argc-2, argv+2, &index);

  if (index == TCL_ERROR) {
    /*
     * oops?
     */
    return TCL_ERROR;
  }

  if (thetree == NULL) {
    /*
     * Are there any constructors at all?
     */
    if (TclObj_DumpMemberNames (interp, theinfo, "new", &index) == TCL_ERROR)
      return TCL_ERROR;

    if (!index) {
      /* nope */
      Tcl_AppendResult (interp, "class \"", theinfo->name,
			"\" does not provide a Tcl-level constructor",
			NULL);
      return TCL_ERROR;
    }

    Tcl_AppendResult (interp, "bad signature: \"", argv[1], "::new", NULL);
    for (index=2; index<argc; index++) {
      if (Tcl_GetCommandInfo (interp, argv[index], &cmdinfo)) {
	if (cmdinfo.clientData) {
	  Tcl_AppendResult (interp, " ",
			    ((Tcl_Object*)cmdinfo.clientData)->classinfo()->name,
			    NULL);
	}
	else {
	  /* huh - does not appear to be an object */
	  Tcl_AppendResult (interp, " *", NULL);
	}
      }
      else {
	/* just a plain string */
	Tcl_AppendResult (interp, " *", NULL);
      }
    }
    
    Tcl_AppendResult      (interp, "\" does not match any of\n", NULL);
    TclObj_DumpSignatures (interp, theinfo, "new", NULL);

    return TCL_ERROR;
  }

  /*
   * okay, go ahead and call the constructor
   */

  if ((thenew =
       (*(theinfo->c_new_proc)) (interp, (clientData)?argv[1]:(char*)NULL,
				 argc, argv, thetree->entry->number))==NULL) {
    /*
     * hmm, the object didn't want to be created. the constructor has
     * already generated a fitting error message
     */
    return TCL_ERROR;
  }

  /*
   * the new operator should call TclObj::TclObj(), which inserts the object
   * into the reverse mapping table. check if this is really true
   */

  sprintf (tmpstr, "%lu", (unsigned long) thenew);
  if ((ptrname = Tcl_GetVar2 (interp, "TclObj_rev",
			      tmpstr, TCL_GLOBAL_ONLY)) == NULL) {
    /*
     * oops?
     */
    sprintf (tmpstr, "%s %s: could not instantiate object",
	     "new", theclass);
    Tcl_SetResult (interp, tmpstr, TCL_VOLATILE);
    delete thenew;
    return TCL_ERROR;
  }
  else {
    /*
     * If this was not a new, create a local variable and a trace on it
     * so that we know when this object must go out of scope
     */
    if (hadold) {
      oldinfo = new Tcl_CmdInfo(cmdinfo);
    }
    else {
      oldinfo = NULL;
    }
    sprintf (tmpstr, "%s_trace", ptrname);
    Tcl_SetVar   (interp, tmpstr, ptrname, 0);
    Tcl_TraceVar (interp, tmpstr, TCL_TRACE_UNSETS,
		  TclObj_scope, oldinfo);
  }

  /*
   * okay
   */

  Tcl_SetResult (interp, ptrname, TCL_VOLATILE);
  return TCL_OK;
}

/*
 * This function is called whenever a local object goes out of scope. This
 * is achieved by using an unset trace on a local variable. This callback
 * then removes the object, and possibly reinstates an outer object that
 * becomes visible again. The Tcl_CmdInfo data is this procedure's
 * clientData
 */

static char *
TclObj_scope (ClientData clientData, Tcl_Interp *interp,
	      char *name, char *name2, int flags)
{
  Tcl_CmdInfo myinfo, *oldinfo = (Tcl_CmdInfo *) clientData;
  char *objname;

  /*
   * At this point, we *know* that this is an unset operation.
   */

  if ((objname = Tcl_GetVar (interp, name, flags&TCL_GLOBAL_ONLY)) != NULL) {
    if (Tcl_GetCommandInfo (interp, objname, &myinfo) == 1) {
      delete (Tcl_Object *) myinfo.clientData;
    }
  }

  /*
   * reinstate previous object
   */
  if (oldinfo) {
    Tcl_CreateCommand (interp, objname,
		       oldinfo->proc,
		       oldinfo->clientData,
		       oldinfo->deleteProc);
    delete oldinfo;
  }

  return NULL;
}

/*
 * delete command
 *  delete "objectname"
 */

static int
TclObj_delete (ClientData clientData, Tcl_Interp *interp,
	       int argc, char *argv[])
{
  Tcl_CmdInfo theinfo;
  int index;

  if (argc < 2) {
    Tcl_AppendResult (interp, "wrong # args: should be \"",
		      argv[0], " object ?object ...?\"", NULL);
    return TCL_ERROR;
  }

  /*
   * just identify the object and call its destructor
   */

  for (index=1; index<argc; index++) {
    if (Tcl_GetCommandInfo (interp, argv[index], &theinfo) == 0) {
      Tcl_AppendResult (interp, "object not found: ", argv[index],
			(index+1<argc)?"\n":"", NULL);
    }
    else {
      Tcl_AppendResult (interp, argv[index], " destroyed", 
			(index+1<argc)?"\n":"", NULL);
      delete (Tcl_Object *) theinfo.clientData;
    }
  }
  return TCL_OK;
}

/*
 * ======================================================================
 * Tcl Initialization for this module
 * ======================================================================
 */

int
Tclobj_Init (Tcl_Interp *interp)
{
  extern TclObj_Info C_Tcl_Object_Info;

  /*
   * register commands
   */
  Tcl_CreateCommand (interp, "new",    TclObj_new,    NULL, NULL);
  Tcl_CreateCommand (interp, "delete", TclObj_delete, NULL, NULL);

  /*
   * Initialize base object
   */
  return TclObj_Register (interp, &C_Tcl_Object_Info);
}

} /* end extern "C" */
