/*
 * Declarations for Tcl_Object. These functions allow properly written
 * C++ classes to be accessible from the Tcl level.
 *
 * We declare the abstract base class, Tcl_Object, from which all other
 * classes that want to be accessible must be derived. Then we define
 * a couple of macros and inline functions to make the developer's life
 * slightly easier, and also declare some helper functions that are of
 * more or less value to the general programmer.
 *
 * $Id: tclobj.h,v 1.3 1996/11/28 15:12:15 fp Exp $
 */

#ifndef __TCLOBJ_H__
#define __TCLOBJ_H__

#include <tcl.h>
#include <stdlib.h>

/*
 * forward decl
 */

class Tcl_Object;

/*
 * command table
 */

typedef struct {
  int number;		/* number of this command                        */
  char *name;		/* name of this command, "new" for constructors  */
  char *args;		/* argument list for this command                */
  char *desc;		/* optional plaintext description                */
} TclObj_Comtab;

/*
 * Info about a class. Each class exports a structure of this type; it
 * is built automatically from the arguments to TCL_OBJECT_DECL_BODY.
 * The base class, Tcl_Object, is no exception -- to most functions, it
 * is a class like any other. The only difference is that the parent
 * list is NULL for Tcl_Object.
 *
 * The structure must be static for each module, it is extensively
 * pointed to. The members "iter", "compats", "members" and "counter"
 * must be NULL in the original structure, they are filled in during
 * initialization (or, as I put it, registration).
 */

struct _TclObj_Signature;
struct _TclObj_Comptree;

typedef struct _TclObj_Info {
  unsigned long               ident;	/* assigned class identifier     */
  const char *                name;	/* static ptr: my name           */
  const char *                parent;	/* class I am derived from       */
  const TclObj_Comtab *       comtab;	/* static ptr: my comtable       */
  
  struct _TclObj_Comptree  *  compats;	/* compatibility tree            */
  struct _TclObj_Signature *  members;	/* signatures of member funcs    */

  unsigned long               counter;	/* counter for next ident no     */

  int          (*c_init_proc) (Tcl_Interp *);	     /* custom init proc */
  Tcl_Object * (*c_new_proc)  (Tcl_Interp *,	         /* new operator */
			       const char * const,
			       int, char *[], int);
  int          (*c_tcl_proc)  (Tcl_Object *,            /* tcl interface */
			       Tcl_Interp *,
			       int, char *[], int);
} TclObj_Info;

/*
 * ----------------------------------------------------------------------
 * class definition of the abstract base class
 * ----------------------------------------------------------------------
 */

class Tcl_Object {
public:
  Tcl_Object (void);
  virtual ~Tcl_Object  ();

  /*
   * Tcl_Enable
   *    Renders an object visible on the Tcl level. No need to call
   *    it yourself -- it is implicitely called wherever necessary.
   * Tcl_Name
   *    Returns the name we have on the Tcl level.
   */

  const char * Tcl_Enable  (Tcl_Interp *, const char * = NULL);
  void         Tcl_Disable (void);
  const char * Tcl_Name    (Tcl_Interp *);

  /*
   * Provide the info function
   */

  int Tcl_Proc (Tcl_Interp *,
		int, char *[],
		int);

  /*
   * Return info about this object
   */

  virtual TclObj_Info * classinfo  (void) const;

  /*
   * The global interpreter
   */

  Tcl_Interp * interpreter;

private:
  const char * ourclasstype;
  static TclObj_Info * Tcl_Object_myinfo;
};

/*
 * Constructor and Destructor
 */

inline
Tcl_Object::Tcl_Object (void)
{
  interpreter  = NULL;
  ourclasstype = NULL;
}

/*
 * Destructor: Disable this object
 */

inline
Tcl_Object::~Tcl_Object (void)
{
  Tcl_Disable();
}

/*
 * definition of inline functions
 */

inline TclObj_Info *
Tcl_Object::classinfo (void) const
{
  return Tcl_Object_myinfo;
}

/*
 * ----------------------------------------------------------------------
 * Export the helper functions from interface.cc
 * ----------------------------------------------------------------------
 */

Tcl_Object *	TclObj_Cast		(Tcl_Interp *,
					 Tcl_Object *,
					 unsigned long oident);

Tcl_Object *	TclObj_Cast		(Tcl_Interp *,
					 const char *,
					 const char * = NULL);

TclObj_Info *	TclObj_Require		(Tcl_Interp *,
					 const char *);

int		TclObj_Register		(Tcl_Interp *,
					 TclObj_Info *);

/*
 * Cast lots of arguments and return lots of result values
 */

int		TclObj_GetInt		(Tcl_Interp *,
					 const char *,
					 int *);

int		TclObj_GetLong		(Tcl_Interp *,
					 const char *,
					 long *);

int		TclObj_GetUnsignedInt	(Tcl_Interp *,
					 const char *,
					 unsigned int *);

int		TclObj_GetUnsignedLong	(Tcl_Interp *,
					 const char *,
					 unsigned long *);

int		TclObj_GetPointer	(Tcl_Interp *,
					 const char *,
					 void **);

int		TclObj_CastArgs		(Tcl_Interp *,
					 int, char *[],
					 const char *, ...);

int		TclObj_MkResult		(Tcl_Interp *,
					 const char *, ...);

/*
 * ----------------------------------------------------------------------
 * Cast arguments to various types
 * 
 * The cast to an object must be a macro because we need the stringified
 * type, the others can be can be inlines, which have the advanatage that
 * they can have local variables and can be called with expressions with
 * side effects as parameter.
 *
 * These macros and functions perform *no* error checking, but that isn't
 * necessary in your Tcl_Proc functions because if you got there, the
 * arguments have already checked out okay.
 * ----------------------------------------------------------------------
 */

#define TCL_OBJECT_ARG_OBJECT(interp,arg,type) \
     ((type *) TclObj_Cast (interp, arg, #type))
#define TCL_OBJECT_ARG_STRING(interp,arg) (arg)

inline int
TCL_OBJECT_ARG_INT (Tcl_Interp *interp, const char *arg)
{
  int result;
  TclObj_GetInt (interp, (char *) arg, &result);
  return result;
}

inline double
TCL_OBJECT_ARG_DOUBLE (Tcl_Interp *interp, const char *arg)
{
  double result;
  Tcl_GetDouble (interp, (char *) arg, &result);
  return result;
}

inline int
TCL_OBJECT_ARG_BOOL (Tcl_Interp *interp, const char *arg)
{
  int result;
  Tcl_GetBoolean (interp, (char *) arg, &result);
  return result;
}

inline long
TCL_OBJECT_ARG_LONG (Tcl_Interp *interp, const char *arg)
{
  long result;
  TclObj_GetLong (interp, arg, &result);
  return result;
}

inline unsigned long
TCL_OBJECT_ARG_ULONG (Tcl_Interp *interp, const char *arg)
{
  unsigned long result;
  TclObj_GetUnsignedLong (interp, arg, &result);
  return result;
}

inline unsigned int
TCL_OBJECT_ARG_UINT (Tcl_Interp *interp, const char *arg)
{
  unsigned int result;
  TclObj_GetUnsignedInt (interp, arg, &result);
  return result;
}

inline void *
TCL_OBJECT_ARG_PTR (Tcl_Interp *interp, const char *arg)
{
  void *result;
  TclObj_GetPointer (interp, arg, &result);
  return result;
}

/*
 * ----------------------------------------------------------------------
 * Return some values
 * ----------------------------------------------------------------------
 */

inline int
TCL_OBJECT_RETURN_OBJECT (Tcl_Interp *interp, Tcl_Object *obj)
{
  const char *name = obj->Tcl_Enable (interp);
  Tcl_SetResult (interp, (char *) name, TCL_STATIC);
  return TCL_OK;
}

inline int
TCL_OBJECT_RETURN_INT (Tcl_Interp *interp, int val)
{
  char tmpstr[32];
  sprintf (tmpstr, "%d", val);
  Tcl_SetResult (interp, tmpstr, TCL_VOLATILE);
  return TCL_OK;
}

inline int
TCL_OBJECT_RETURN_DOUBLE (Tcl_Interp *interp, double val)
{
  char tmpstr[32];
  sprintf (tmpstr, "%f", val);
  Tcl_SetResult (interp, tmpstr, TCL_VOLATILE);
  return TCL_OK;
}

inline int
TCL_OBJECT_RETURN_BOOL (Tcl_Interp *interp, int val)
{
  char tmpstr[32];
  sprintf (tmpstr, "%s", val?"true":"false");
  Tcl_SetResult (interp, tmpstr, TCL_VOLATILE);
  return TCL_OK;
}

inline int
TCL_OBJECT_RETURN_STRING (Tcl_Interp *interp, const char *str)
{
  Tcl_SetResult (interp, (char *) str, TCL_VOLATILE);
  return TCL_OK;
}

inline int
TCL_OBJECT_RETURN_LONG (Tcl_Interp *interp, long val)
{
  char tmpstr[32];
  sprintf (tmpstr, "%ld", val);
  Tcl_SetResult (interp, tmpstr, TCL_VOLATILE);
  return TCL_OK;
}

inline int
TCL_OBJECT_RETURN_UINT (Tcl_Interp *interp, unsigned int val)
{
  char tmpstr[32];
  sprintf (tmpstr, "%u", val);
  Tcl_SetResult (interp, tmpstr, TCL_VOLATILE);
  return TCL_OK;
}

inline int
TCL_OBJECT_RETURN_ULONG (Tcl_Interp *interp, unsigned long val)
{
  char tmpstr[32];
  sprintf (tmpstr, "%lu", val);
  Tcl_SetResult (interp, tmpstr, TCL_VOLATILE);
  return TCL_OK;
}

inline int
TCL_OBJECT_RETURN_PTR (Tcl_Interp *interp, void *val)
{
  char tmpstr[32];
  sprintf (tmpstr, "pv!%lu", (unsigned long) val);
  Tcl_SetResult (interp, tmpstr, TCL_VOLATILE);
  return TCL_OK;
}

/*
 * ----------------------------------------------------------------------
 * Definitions for derived classes
 * ----------------------------------------------------------------------
 */

/*
 * declarations within the class definition
 */

#define TCL_OBJECT_DECL_CLASS(the_class) \
private: \
  static TclObj_Info * the_class ## _myinfo; \
public: \
  TclObj_Info * classinfo (void) const { return the_class ## _myinfo; }

/*
 * Declarations in the .cc file
 */

#define TCL_OBJECT_DECL_BODY(the_class,parent,comtable,init) \
extern "C" { \
Tcl_Object * \
the_class ## _New (Tcl_Interp *interp, const char * const myname, \
		   int argc, char *argv[], int command) \
{ \
  parent *myself = new the_class (interp, argc, argv, command); \
  myself->Tcl_Enable (interp, myname); \
  return myself; \
} \
int \
the_class ## _Tcl_Proc (Tcl_Object *myself, Tcl_Interp *interp, \
			int argc, char *argv[], int command) \
{ \
  return ((the_class *)(parent *)myself)->Tcl_Proc (interp, argc, argv, \
						    command); \
} \
\
TclObj_Info C_ ## the_class ## _Info = { \
  0, #the_class, #parent, comtable, NULL, NULL, \
  0, init, the_class ## _New, the_class ## _Tcl_Proc \
}; \
\
TclObj_Info * the_class ## :: ## the_class ## _myinfo = \
     & C_ ## the_class ## _Info; \
\
extern "C" int C_ ## parent ## _Init (Tcl_Interp *); \
\
int \
C_ ## the_class ## _Init (Tcl_Interp *interp) \
{ \
  if (C_ ## parent ## _Init (interp) == TCL_ERROR) return TCL_ERROR; \
  return TclObj_Register (interp, & C_ ## the_class ## _Info); \
} \
}

/*
 * Declarations in the .cc file for an abstract class, which does not
 * export a new operator.
 */

#define TCL_OBJECT_DECL_ABSTRACT_BODY(the_class,parent,comtable,init) \
extern "C" { \
int \
the_class ## _Tcl_Proc (Tcl_Object *myself, Tcl_Interp *interp, \
			int argc, char *argv[], int command) \
{ \
  return ((the_class *) myself)->Tcl_Proc (interp, argc, argv, command); \
} \
\
TclObj_Info C_ ## the_class ## _Info = { \
  0, #the_class, #parent, comtable, NULL, NULL, \
  0, init, NULL, the_class ## _Tcl_Proc \
}; \
\
TclObj_Info * the_class ## :: ## the_class ## _myinfo = \
     & C_ ## the_class ## _Info; \
\
extern "C" int C_ ## parent ## _Init (Tcl_Interp *); \
\
int \
C_ ## the_class ## _Init (Tcl_Interp *interp) \
{ \
  if (C_ ## parent ## _Init (interp) == TCL_ERROR) return TCL_ERROR; \
  return TclObj_Register (interp, & C_ ## the_class ## _Info); \
} \
}

/*
 * ----------------------------------------------------------------------
 * Some helpers
 * ----------------------------------------------------------------------
 */

/*
 * This macro can be called in the default branch of the Tcl_Proc
 * procedure. It produces an error message about the wrong command
 * This should never be necessary, anyway.
 */

#define TCL_OBJECT_NO_MATCH(interp)	{ \
  Tcl_AppendResult (interp, argv[0], ": bad action \"", argv[1], "\" ", \
		    "for object of type ", classinfo()->name, " (oops?) ", NULL); \
  return TCL_ERROR; \
}

/*
 * ----------------------------------------------------------------------
 * Make a package. Usually, you can dynamically load classes using
 * the load command by prepending a c_ to the class name. This doesn't
 * work if one library defines multiple classes
 * ----------------------------------------------------------------------
 */

#define TCL_OBJECT_BEGIN_PACKAGE(name) \
extern "C" int \
Pkg_ ## name ## _Init (Tcl_Interp *interp) \
{

#define TCL_OBJECT_PKG_CLASS(class) \
  extern int C_ ## class ## _Init (Tcl_Interp *); \
  if (C_ ## class ## _Init (interp) == TCL_ERROR) return TCL_ERROR;

#define TCL_OBJECT_END_PACKAGE() \
  return TCL_OK; \
}

#endif
